from ..estimate import cal_assoc, cal_cond_assoc

__all__ = ["cal_general_assoc"]


def cal_general_assoc(x, y, z, method, xtype, ytype, ztype, **kwargs):
    """关联和条件关联的通用计算"""    
    return cal_assoc(x, y, method, xtype, ytype, **kwargs) if z is None \
        else cal_cond_assoc(x, y, z, method, xtype, ytype, ztype, **kwargs)


# -*- coding: utf-8 -*-
"""
Created on 2022/09/18 23:21:37

@File -> mi_cmi.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: MIC和RMIC
"""

import numpy as np

from ...util import SuperCategorEncoding, discretize_arr
from ._mic_rmic import MaximalInfoCoeff, RefinedMaximalInfoCoeff


# ---- 数据压缩编码 ---------------------------------------------------------------------------------

def _transf_labels2char(s: np.ndarray):
    """哎, numpy直接匹配函数没找到, 只能暂时转成字符串来处理了"""
    return "".join(f"-{int(p)}" for p in s)


def _convert_arr2int(arr: np.ndarray):
    """将多维数据按照标签进行编码为一维整数
    :param arr: 具有离散整数值的多维数组
    """
    combines = [tuple(row) for row in arr]
    labels = np.unique(combines, axis=0)
    labels = np.apply_along_axis(_transf_labels2char, 1, labels)
    arr_labels = np.apply_along_axis(_transf_labels2char, 1, arr).reshape(-1, 1)
    return np.apply_along_axis(lambda s: np.where(labels == s)[0][0], 1, arr_labels)


def _reencode(arr: np.ndarray):
    """重新将多维数组编码为1维数组"""
    arr = arr.copy()
    arr = discretize_arr(arr, n = arr.shape[0] // 10, method="qcut")
    arr = _convert_arr2int(arr)
    return arr


class MIC(object):
    """最大信息系数"""
    
    def __init__(self, x, y):
        self.x = x.copy().reshape(len(x), -1)
        self.y = y.copy().reshape(len(y), -1)
    
    def __call__(self, method="rmic", encode=True):
        # 多维数组逐列离散化, 并合并编码压缩为一维数组
        y = _reencode(self.y) if self.y.shape[1] > 1 else self.y.copy()
        x = _reencode(self.x) if self.x.shape[1] > 1 else self.x.copy()
        if method != "rmic":
            return MaximalInfoCoeff(x, y).cal_assoc()
        
        if encode:
            x = SuperCategorEncoding(x, y).encode(method="mhg") # 进行有监督编码, NOTE x值需为int
            
        return RefinedMaximalInfoCoeff(x, y).cal_assoc()
            
            
class CMIC(object):
    """条件最大信息系数"""
    
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
        
    def __call__(self, method="rmic"):
        xy = np.c_[self.x, self.y]
        return MIC(self.x, self.y)(method) + MIC(xy, self.z)(method) - MIC(self.x, self.z)(method) - \
            MIC(self.y, self.z)(method)


# -*- coding: utf-8 -*-
"""
Created on 2023/01/05 20:09:48

@File -> kde.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 基于KDE的信息熵和互信息估计, 仅适用于连续变量
"""

from scipy.stats import gaussian_kde
from numpy import log
import numpy as np

from ...setting import BASE
from ...util import stdize_values


# ---- 数据信息估计 ---------------------------------------------------------------------------------

class MargEntropy(object):
    """边际熵"""
    # NOTE: 仅适用于连续变量
    
    def __init__(self, x: np.ndarray):
        self.x_norm = stdize_values(x, "c")  # shape = (N, dim)
        self.dim = self.x_norm.shape[1]
        
    def __call__(self):
        dens = gaussian_kde(self.x_norm.T)(self.x_norm.T)  # shape = (N,)
        return np.mean(-log(dens) / log(BASE))
    
    
class MutualInfoKDE(object):
    """互信息"""
    # NOTE: 仅适用于连续变量
    
    def __init__(self, x: np.ndarray, y: np.ndarray):
        self.x_norm = stdize_values(x, "c")  # shape = (N, dim)
        self.y_norm = stdize_values(y, "c")
        self.xy_norm = np.c_[self.x_norm, self.y_norm]
        self.dim_x = self.x_norm.shape[1]
        self.dim_y = self.y_norm.shape[1]
        
    def __call__(self):
        dens_xy = gaussian_kde(self.xy_norm.T)(self.xy_norm.T)  # shape = (N,)
        dens_x = gaussian_kde(self.x_norm.T)(self.x_norm.T)  # shape = (N,)
        dens_y = gaussian_kde(self.y_norm.T)(self.y_norm.T)  # shape = (N,)
        return np.mean(log(dens_xy / (dens_x * dens_y)) / log(BASE))


from pyitlib import discrete_random_variable as drv
import pandas as pd
import numpy as np

from ...util import stdize_values


def discretize_series(x: np.ndarray, n: int = None, method="qcut"):
    """对数据序列采用等频分箱"""
    n = len(x) // 15 if n is None else n
    q = int(len(x) // n)

    # 分箱方式.
    if method == "qcut":
        x_enc = pd.qcut(x.flatten(), q, labels=False, duplicates="drop").flatten()  # 等频分箱
    if method == "cut":
        x_enc = pd.cut(x.flatten(), q, labels=False, duplicates="drop").flatten()  # 等宽分箱
        
    return x_enc


class MutualInfoClassic(object):
    """基于经典离散化的互信息和条件互信息计算"""
    
    def __init__(self, x: np.ndarray, y: np.ndarray, z: np.ndarray=None):
        self.x_norm = stdize_values(x, "c")
        self.y_norm = stdize_values(y, "c")
        self.z_norm = stdize_values(z, "c") if z is not None else None
    
    def __call__(self, method="qcut"):
        x_enc = discretize_series(self.x_norm, method=method).astype(int)
        y_enc = discretize_series(self.y_norm, method=method).astype(int)
        if self.z_norm is None:
            mi = drv.information_mutual(x_enc, y_enc)
            return float(mi)
        
        z_enc = discretize_series(self.z_norm, method=method).astype(int)
        cmi = drv.information_mutual_conditional(x_enc, y_enc, z_enc)
        return float(cmi)  # NOTE: drv.infor_mutual_cond居然返回array格式
    

# if __name__ == "__main__":
    
#     def test_cc():
#         x = np.random.normal(0, 1, 100000)
#         y = np.random.normal(0, 1, 100000)
#         print(f"mi = {MutualInfoClassic(x, y)()}")
        
#     def test_ccc():
#         x = np.random.normal(0, 1, 100000)
#         y = np.random.normal(0, 1, 100000)
#         z = np.random.normal(0, 1, 100000)
#         print(f"cmi = {MutualInfoClassic(x, y, z)()}")
        
#     test_cc()
#     test_ccc()


# -*- coding: utf-8 -*-
# sourcery skip: avoid-builtin-shadow
"""
Created on 2023/04/26 16:45:20

@File -> surrog_indep_test.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 基于代用数据的独立性测试
"""

__doc__ = """
    计算关联系数并通过代用数据重采样获得结果的显著性信息
"""

import logging
import random
import numpy as np

from . import cal_general_assoc


def exec_surrog_indep_test(x, y, method, z=None, xtype=None, ytype=None, ztype=None, rounds=100, 
                           alpha=0.05, max_size_bt=1000, size_bt=None, **kwargs):
    """执行基于代用数据的独立性检验

    :param x: x数据, np.ndarray
    :param y: y数据, np.ndarray
    :param method: 选用的关联度量方法, str
    :param z: z数据, np.ndarray, defaults to None
    :param xtype: x的数值类型, str, defaults to None
    :param ytype: y的数值类型, str, defaults to None
    :param ztype: y的数值类型, str, defaults to None
    :param rounds: 重复轮数, defaults to 100
    :param alpha: 显著性阈值, defaults to 0.05
    :param max_size_bt: 用于自助重采样的最大样本数, defaults to 1000
    :param size_bt: 自行设定的重采样规模, int, defaults to None
    :param kwargs: cal_general_assoc方法中的关键字参数
    :return
        assoc: 关联值, float
        p: P值, float
        indep: 是否独立, bool
        assocs_srg: 代理数据关联值数组, np.ndarray
    """
    # 如果样本量小, 则全部代入计算; 否则, 选择max_size_bt的样本进行计算
    if size_bt is None:
        if len(x) <= max_size_bt:
            size_bt = len(x)
        else:
            logging.warning(
                f"exec_surrog_indep_test: 采用默认max_size_bt={max_size_bt}作为size_bt, \
                用于限制代用数据重采样规模")
            size_bt = max_size_bt
        
    # 计算关联系数
    # TODO: 通过随机抽样获得关联值分布, 有放回or无放回?
    if z is None:
        assoc = cal_general_assoc(x[:size_bt], y[:size_bt], z, method, xtype, ytype, ztype, **kwargs)
    else:
        assoc = cal_general_assoc(x[:size_bt], y[:size_bt], z[:size_bt], method, xtype, ytype, ztype, **kwargs)

    # 计算背景值
    idxs = np.arange(len(x))
    assocs_srg = np.array([])
    for _ in range(rounds):
        # 基于代用数据获得背景值
        idxs_bt = random.choices(idxs, k=size_bt)  # NOTE: 有放回抽样
        x_bt, y_bt = x[idxs_bt], y[idxs_bt]
        x_srg = np.random.permutation(x_bt)  # 随机重排获得代用数据
        if z is None:
            assoc_srg = cal_general_assoc(x_srg, y_bt, None, method, xtype, ytype, ztype, **kwargs)
        else:
            z_bt = z[idxs_bt]
            assoc_srg = cal_general_assoc(x_srg, y_bt, z_bt, method, xtype, ytype, ztype, **kwargs)

        assocs_srg = np.append(assocs_srg, assoc_srg)

    # 计算显著性
    p = len(assocs_srg[assocs_srg >= assoc]) / rounds
    indep = p >= alpha
    return assoc, (p, indep, assocs_srg)


# sourcery skip: avoid-builtin-shadow
from joblib import Parallel, delayed
from typing import List
import numpy as np
import random

from ..estimate import cal_assoc, cal_cond_assoc
from ..util import stdize_values

EPS = 1e-12
MI_METHOD = "MI-GIEF"
CMI_METHOD = "CMI-GIEF"

__doc__ = """
    基于CMIM-GIEF的紧凑关联变量识别:
    NOTE:
    1. 在计算中将所有变量视为连续类型, 经过stdize_value处理为标准格式
"""


# ---- 关联系数计算 ---------------------------------------------------------------------------------

def cal_mi(x, y, k, n):
    """计算I(x;y)"""
    x, y = x.flatten(), y.flatten()
    idxs = random.sample(range(x.shape[0]), k=n)    # 随机无返回抽样
    x, y = x[idxs], y[idxs]
    return cal_assoc(x, y, MI_METHOD, "c", "c", k=k)


def cal_cmi(xi, xs, y, k, n):
    """计算I(xi;y|xs)"""
    xi, xs, y = xi.flatten(), xs.flatten(), y.flatten()
    idxs = random.sample(range(xi.shape[0]), k=n)   # 随机无返回抽样
    xi, xs, y = xi[idxs], xs[idxs], y[idxs]
    return cal_cond_assoc(xi, y, xs, CMI_METHOD, "c", "c", "c", k=k)


# ---- CMIM-GIEF -----------------------------------------------------------------------------------

class CondMIMaximization(object):
    """条件关联系数最大化"""
    
    def __init__(self, X: np.ndarray, y: np.ndarray, n_jobs: int = 2, k: int = 3, n: int=500):
        """初始化"""
        # 将所有变量均视为连续值类型, 加入噪声并归一化
        self.X = stdize_values(X, "c", EPS)
        self.y = stdize_values(y, "c", EPS)
        self.N, self.D = self.X.shape
        self.n_jobs = n_jobs
        self.k = k      # MI-GIEF中计算的K近邻数
        self.n = n      # 用于限制MI-GIEF和CMI-GIEF计算样本量
        
    def _cal_mi(self, x, y):
        return cal_mi(x, y, self.k, self.n)
    
    def _cal_cmi(self, xi, xs, y):
        return cal_cmi(xi, xs, y, self.k, self.n)
        
    def _init_assocs(self) -> List[float]:
        """计算所有特征与目标的关联系数"""
        with Parallel(n_jobs=self.n_jobs) as parallel:
            assocs = parallel(delayed(self._cal_mi)(self.X[:, i], self.y) for i in range(self.D))
        return assocs
    
    def _recog_redundants(self, S: set, F: set, cmi_thres=0.1) -> set:
        R_new = set()
        for fi in F.difference(S):
            for fs in S:
                cassoc = self._cal_cmi(self.X[:, fi], self.X[:, fs], self.y)
                if cassoc < cmi_thres:
                    R_new.update({fi})
                    break
        return R_new
    
    def feature_selection(self, S_pre: set = None, max_iters: int = 30, verbose = False):
        """特征选择
        :param S_pre: 指定X中已被提前选中的特征集合, defaults to None
        :param max_iters: 迭代选择的最大次数, defaults to 30
        :param verbose: 是否打印过程输出, defaults to False
        """
        F = set(range(self.D))  # 全量特征集合
        S = set() if S_pre is None else S_pre.copy()
        R = set()
        features_rank = []  # 用于记录所选特征顺序

        # 识别冗余特征集R
        if S:
            R.update(self._recog_redundants(S, F))

        cmi_matrix = None                   # 用于记录迭代中的CMI值
        min_cmi = np.ones(len(F)) * np.nan  # 用于记录迭代中的最小CMI值
        i = 0
        while True:
            C = F.difference(S).difference(R)  # 剩余候选特征

            if (not C) or (len(features_rank) > max_iters):
                break
            if not S:  # 如果一个被选的特征都没有, 则采用单因素分析中所得选择关联度最强的
                self.assocs = self._init_assocs()
                f = np.argmax(self.assocs)
            else:
                if cmi_matrix is None:
                    cmi_matrix = np.ones((len(F), len(F))) * np.nan
                    for fi in C:
                        xi = self.X[:, fi]
                        for fs in S:
                            xs = self.X[:, fs]
                            cmi_matrix[fi, fs] = self._cal_cmi(xi, xs, self.y)
                        min_cmi[fi] = np.nanmin(cmi_matrix[fi, :])
                else:
                    xf = self.X[:, f]
                    for fi in C:
                        xi = self.X[:, fi]
                        cmi_matrix[fi, f] = self._cal_cmi(xi, xf, self.y)
                        min_cmi[fi] = np.nanmin([min_cmi[fi], cmi_matrix[fi, f]])

                f = list(C)[np.nanargmax(min_cmi[list(C)])]
            
            R_new = self._recog_redundants({f}, F.difference(S))
            S.update({f})           # 加入新选择的特征f
            R.update(R_new)         # 增加候选特征中关于f的冗余
            features_rank.append(f)
            
            if verbose:
                print(f"iteration i={i} \t select feature no. {f}")
            
            i += 1
            if i == max_iters:
                break
            
        details = (F, S, R, cmi_matrix, min_cmi)
        return features_rank, details


import numpy as np

from ..util import exec_model_test


# 紧凑关联识别
def recog_compact(X, y, features_rank, model, n_features_lst, S_pre: set, metric: str="r2", 
                  rounds: int=3):
    """从CMIM-GIEF所得features_rank中识别出y的紧凑关联变量, 返回其在X中的编号表

    :param X: X数组, np.ndarray
    :param y: y数组, np.ndarray
    :param features_rank: CMIM-GIEF所得特征排序表, list
    :param model: 用于进行FIMT测试的机器学习模型
    :param n_features_lst: 用于FIMT测试的特征数列表
    :param S_pre: 在X中已被事先选中的特征集, set
    :param metric: 模型评估指标, str, defaults to "r2"
    :param rounds: 每个特征集合上重复评估的轮数, defaults to 3
    """          

    # 当前仅支持基于决定系数的判断
    if metric != "r2":
        raise ValueError(f"unsupported metric {metric}")

    if S_pre:
        init_value = exec_model_test(X[:, list(S_pre)], y, model, metric=metric, rounds=rounds)[0]
    else:
        init_value = 0.0  # 仅考虑已选特征所得模型分数

    # FIMT测试
    fimt_values = []
    for n_features in n_features_lst:
        S = list(S_pre) + features_rank[:n_features]
        fimt_values.append(exec_model_test(X[:, S], y, model, metric=metric, rounds=rounds)[0])

    # 截断点判断
    max_value = np.max(fimt_values)

    # 如果FIMT测试最大值不高于已选特征集对应的初始值, 则不新选入
    best_n_features = 0
    if max_value - init_value >= 0.01:
        for i, n_features in enumerate(n_features_lst):
            if fimt_values[i] > max_value - 0.01:
                best_n_features = n_features
                break

    # 新识别出的紧凑变量在X所有特征中的序号
    features_compact = features_rank[:best_n_features]
    return features_compact, fimt_values


import warnings

warnings.filterwarnings("ignore")

from sklearn.neighbors import BallTree, KDTree
from sklearn.preprocessing import MinMaxScaler
from sklearn.metrics import mean_squared_error as mse, explained_variance_score as evs,\
    r2_score as r2
from scipy.special import gamma
import category_encoders as ce
import pandas as pd
import numpy as np
import random


####################################################################################################
# 数据编码
####################################################################################################

UNSUPER_METHODS = ["ordinal", "random", "count"]
SUPER_METHODS = ["target", "m_estimator", "james_stein", "glmm", "woe", "leave_one_out", "catboost", "mhg"]


class UnsuperCategorEncoding(object):
    """无监督一维类别值编码"""
    
    def __init__(self, x: np.ndarray or list):
        self.x = pd.Series(np.array(x).astype(np.int).flatten(), name = "x")  # type: pd.Series
        self.N = len(x)
        
    def ordinal_encoding(self):
        enc = ce.OrdinalEncoder(cols = ["x"])
        enc.fit(self.x)
        x_enc = enc.transform(self.x)
        return np.array(x_enc).flatten()
    
    def random_encoding(self, seed: int = None):
        # TODO: 优化效率.
        x_sorted = sorted(set(self.x))

        if seed is not None:
            random.seed(seed)
        random.shuffle(x_sorted)

        codes = list(range(1, len(x_sorted) + 1))
        x_enc = self.x.replace(dict(zip(x_sorted, codes)))
        return np.array(x_enc).flatten()
    
    def count_encoding(self):
        enc = ce.CountEncoder(cols = ["x"])
        enc.fit(self.x)
        x_enc = enc.transform(self.x)
        return np.array(x_enc)
    
    def encode(self, method: str, **kwargs):
        try:
            assert method in UNSUPER_METHODS
        except:
            raise ValueError(f'Invalid method = \"{method}\"')
        
        x_enc = None
        if method == "ordinal":
            x_enc = self.ordinal_encoding()
        elif method == "random":
            x_enc = self.random_encoding(**kwargs)
        elif method == "count":
            x_enc = self.count_encoding()
        return x_enc
    

class SuperCategorEncoding(object):
    """有监督一维序列编码
    Example
    ----
    super_enc = SuperCategorEncoding(x, y)
    x_enc = super_enc.mhg_encoding()
    """
    
    def __init__(self, x: np.ndarray or list, y: np.ndarray or list):
        """
        初始化
        :param x: x序列, x一定为Nominal类别型变量
        :param y: y序列, y一定为数值型变量
        """
        # NOTE x值需为int
        self.x = pd.Series(np.array(x).astype(np.int).flatten(), name = "x")  # type: pd.Series
        self.y = pd.Series(np.array(y).astype(np.float32).flatten(), name = "y")  # type: pd.Series

    def target_encoding(self):
        enc = ce.TargetEncoder(cols = ["x"], smoothing = 1.0)
        return self._encode_transform(enc)
    
    def m_estimator_encoding(self):
        enc = ce.MEstimateEncoder(cols = ["x"], m = 20.0)
        return self._encode_transform(enc)
    
    def james_stein_encoding(self):
        enc = ce.JamesSteinEncoder(cols = ["x"])
        return self._encode_transform(enc)
    
    def glmm_encoding(self):
        enc = ce.GLMMEncoder(cols = ["x"])
        return self._encode_transform(enc)
    
    def woe_encoding(self, **kwargs):
        y_mean = self.y.mean()
        y_binar = self.y.copy()
        y_binar[np.array(self.y > y_mean).flatten()] = 1
        y_binar[np.array(self.y <= y_mean).flatten()] = 0

        enc = ce.WOEEncoder(cols = ["x"], **kwargs)
        enc.fit(self.x, y_binar)
        x_enc = enc.transform(self.x)
        return np.array(x_enc).flatten()
    
    def leave_one_out_encoding(self):
        enc = ce.LeaveOneOutEncoder(cols = ["x"])
        return self._encode_transform(enc)
    
    def catboost_encoding(self):
        enc = ce.CatBoostEncoder(cols = ["x"])
        return self._encode_transform(enc)

    def _encode_transform(self, enc):
        """编码并转换"""
        enc.fit(self.x, self.y)
        x_enc = enc.transform(self.x)
        return np.array(x_enc).flatten()
    
    def mhg_encoding(self):
        d = pd.concat([self.x, self.y], axis = 1)
        y_mean = d.groupby("x").mean()
        x_enc = self.x.replace(dict(zip(y_mean.index, list(y_mean["y"]))))
        return np.array(x_enc).flatten()
    
    def encode(self, method: str, **kwargs):
        try:
            assert method in SUPER_METHODS
        except:
            raise ValueError(f'Invalid method = \"{method}\"')
        
        x_enc = None
        if method == "target":
            x_enc = self.target_encoding()
        elif method == "m_estimator":
            x_enc = self.m_estimator_encoding()
        elif method == "james_stein":
            x_enc = self.james_stein_encoding()
        elif method == "glmm":
            x_enc = self.glmm_encoding()
        elif method == "woe":
            x_enc = self.woe_encoding(**kwargs)
        elif method == "leave_one_out":
            x_enc = self.leave_one_out_encoding()
        elif method == "catboost":
            x_enc = self.catboost_encoding()
        elif method == "mhg":
            x_enc = self.mhg_encoding()
        return x_enc
    

####################################################################################################
# 数据处理 
####################################################################################################

# ---- 数据标准化 -----------------------------------------------------------------------------------

def normalize(X):
    scaler = MinMaxScaler()
    X = scaler.fit_transform(X.copy())
    return X


def _convert_1d_series2int(x):
    """将一维数据按照标签进行编码为连续整数"""
    x = x.flatten()
    x_unique = np.unique(x)

    if len(x_unique) > 100:
        raise RuntimeWarning(
            f"too many labels: {len(x_unique)} for the discrete data")

    x = np.apply_along_axis(lambda x: np.where(
        x_unique == x)[0][0], 1, x.reshape(-1, 1))
    return x


def _convert_arr2int(arr):
    """将一维数据按照标签进行编码为连续整数"""
    _, D = arr.shape
    for d in range(D):
        arr[:, d] = _convert_1d_series2int(arr[:, d])
    return arr.astype(int)


def stdize_values(x: np.ndarray, dtype: str, eps: float = 1e-10):
    """数据预处理: 标签值整数化、连续值归一化, 将连续和离散变量样本处理为对应的标准格式用于后续分析"""
    x = x.copy()
    x = x.reshape(x.shape[0], -1)
    if dtype == "c":
        # 连续值加入噪音并归一化
        x += eps * np.random.random_sample(x.shape)
        return normalize(x)
    elif dtype == "d":
        # 将标签值转为连续的整数值
        x = _convert_arr2int(x)
        return x
    

# ---- 数据离散化 -----------------------------------------------------------------------------------

def _discretize_series(x: np.ndarray, n: int = 30, method="qcut"):
    """对数据序列采用等频分箱"""
    q = int(len(x) // n)
    if method == "qcut":
        x_enc = pd.qcut(x, q, labels=False, duplicates="drop").flatten()  # 等频分箱
    elif method == "cut":
        x_enc = pd.cut(x, q, labels=False, duplicates="drop").flatten()  # 等宽分箱
    return x_enc


def discretize_arr(X: np.ndarray, n: int = None, method: str = "qcut"):
    """逐列离散化"""
    if n is None:
        n = X.shape[0] // 20
    X = X.copy()
    for i in range(X.shape[1]):
        X[:, i] = _discretize_series(X[:, i], n, method)
    return X.astype(int)
    

# ---- K近邻查询 ------------------------------------------------------------------------------------
    
def build_tree(x, metric: str = "chebyshev"):
    """建立近邻查询树. 低维用具有欧式距离特性的KDTree; 高维用具有更一般距离特性的BallTree"""
    return BallTree(x, metric=metric) if x.shape[1] >= 20 else KDTree(x, metric=metric)


def query_neighbors_dist(tree: BallTree or KDTree, x, k: int):
    """求得x样本在tree上的第k个近邻样本"""
    return tree.query(x, k=k + 1)[0][:, -1]


# ---- 空间球体积 -----------------------------------------------------------------------------------

def get_unit_ball_volume(d: int, metric: str = "euclidean"):
    """d维空间中按照euclidean或chebyshev距离计算所得的单位球体积"""
    if metric == "euclidean":
        return (np.pi ** (d / 2)) / gamma(1 + d / 2)  
    elif metric == "chebyshev":
        return 1
    else:
        raise ValueError(f"unsupported metric {metric}")
    

# ---- 构造时延序列 ---------------------------------------------------------------------------------

def build_td_series(x: np.ndarray, y: np.ndarray, tau: int, max_len: int = 5000):
    """
    生成时延序列样本
    :param x: 样本x数组
    :param y: 样本y数组
    :param tau: 时间平移样本点数, 若 tau > 0, 则 x 对应右方 tau 个样本点后的 y; 
        若 tau < 0, 则 y 对应右方 tau 个样本点后的 x
    """
    x_td, y_td = x.flatten(), y.flatten()
    
    if len(x_td) != len(y_td):
        raise ValueError("length of x is not equal to y")
    
    N = len(x_td)
    
    lag_remain = np.abs(tau) % N
    if lag_remain != 0:
        if tau > 0:
            y_td = y_td[lag_remain:]
            x_td = x_td[:-lag_remain]
        else:
            x_td = x_td[lag_remain:]
            y_td = y_td[:-lag_remain]

    # 当样本量过高时执行降采样, 降低计算复杂度
    if len(x_td) > max_len:
        idxs = np.arange(len(x_td))
        np.random.shuffle(idxs)
        idxs = idxs[:max_len]
        x_td, y_td = x_td[idxs], y_td[idxs]

    return x_td, y_td


####################################################################################################
# 模型评估 
####################################################################################################

def _cal_metric(y_true, y_pred, metric: str):
    y_true, y_pred = y_true.flatten(), y_pred.flatten()
    if metric == "r2":
        return r2(y_true, y_pred)
    if metric == "evs":
        return evs(y_true, y_pred)
    if metric == "mse":
        return mse(y_true, y_pred)
    if metric == "mape":
        idxs = np.where(y_true != 0)
        y_true = y_true[idxs]
        y_pred = y_pred[idxs]
        return np.sum(np.abs((y_pred - y_true) / y_true)) / len(y_true)


# 模型评价
def exec_model_test(X, y, model, metric: str="r2", test_ratio: float=0.3, rounds: int=10):
    """执行建模测试"""
    X, y = X.copy(), y.copy()
    N = X.shape[0]
    test_size = int(N * test_ratio)
    metrics = []
    for _ in range(rounds):
        shuffled_indexes = np.random.permutation(range(N))
        train_idxs = shuffled_indexes[test_size:]
        test_idxs = shuffled_indexes[:test_size]

        X_train, X_test = X[train_idxs, :], X[test_idxs, :]
        y_train, y_test = y[train_idxs], y[test_idxs]

        model.fit(X_train, y_train)
        m = _cal_metric(y_test, model.predict(X_test), metric)
        metrics.append(m)
    return np.mean(metrics), metrics


# -*- coding: utf-8 -*-
"""
Created on 2022/09/18 16:32:51

@File -> cmi.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 条件互信息
"""

import numpy as np

from ....util import stdize_values
from ....setting import DTYPES, BASE
from ..entropy.cond_entropy import CondEntropy
from ..entropy.marg_entropy import MargEntropy
from ._kraskov import kraskov_mi as cal_kraskov_cmi
from .mi import MutualInfoGIEF


class CondMutualInfoGIEF(object):
    """条件互信息"""
    
    def __init__(self, x: np.ndarray, xtype: str, y: np.ndarray, ytype: str, z: np.ndarray, ztype: str):
        assert xtype in DTYPES
        assert ytype in DTYPES
        assert ztype in DTYPES
        self.x = x.copy()
        self.y = y.copy()
        self.z = z.copy()
        self.xtype = xtype
        self.ytype = ytype
        self.ztype = ztype
        
    def __call__(self, k=3, metric="chebyshev", method: str=None):
        # 如果指定使用Kraskov方法, 或者三个变量数值类型均为"c", 则返回Kraskov熵
        cond_0 = method=="Kraskov"
        cond_1 = (self.xtype=="c") & (self.ytype=="c") & (self.ztype=="c")
        if cond_0 | cond_1:
            x, y, z = stdize_values(self.x, self.xtype), stdize_values(self.y, self.ytype), \
                stdize_values(self.z, self.ztype)
            return cal_kraskov_cmi(x, y, z, k=k, base=BASE, alpha=0)
        
        # 根据z与x或y类型差异选择计算方法
        if self.ztype in [self.xtype, self.ytype]:
            if self.ztype == self.xtype:
                x2, x1 = self.x, self.y
                x2_type, x1_type = self.xtype, self.ytype
            else:
                x2, x1 = self.y, self.x
                x2_type, x1_type = self.ytype, self.xtype
            
            arr = np.c_[x2, self.z]
            return MutualInfoGIEF(arr, x2_type, x1, x1_type)(k, metric) - \
                MutualInfoGIEF(x1, x1_type, self.z, self.ztype)(k, metric)
        else:
            a1 = MutualInfoGIEF(self.x, self.xtype, self.y, self.ytype)(k, metric)
            a2 = CondEntropy(self.z, self.ztype, self.x, self.xtype)(k, metric)
            a3 = CondEntropy(self.z, self.ztype, self.y, self.ytype)(k, metric)
            
            arr = np.c_[self.x, self.y]
            a4 = CondEntropy(self.z, self.ztype, arr, self.xtype)(k, metric)
            a5 = MargEntropy(self.z, self.ztype)(k, metric)
            return a1 + a2 + a3 - a4 - a5


from setuptools import setup, find_packages
from pathlib import Path

VERSION = "0.0.9"
DESCRIPTION = "Package for information estimation, independence test, etc."
this_directory = Path(__file__).parent
long_description = (this_directory/"README.md").read_text(encoding="utf-8")

setup(
    name="giefstat",
    version=VERSION,
    author="Dreisteine",
    author_email="dreisteine@163.com",
    description=DESCRIPTION,
    packages=find_packages(include=["giefstat", "giefstat.*"]),  # NOTE: 注意giefstat.*的写法
    install_requires=["scikit-learn==0.24.0", "minepy==1.2.6"],
    keywords=["python", "information estimation"],
    python_requires=">=3.8.0",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ulti-Dreisteine/general-information-estimation-framework",
)


from minepy import MINE
import numpy as np

ALPHA = 0.6
C = 5
EPS = 1e-6


class MaximalInfoCoeff(object):
    """MIC成对检验"""

    def __init__(self, x: np.ndarray, y: np.ndarray, mic_params: dict = None):
        self.x, self.y = x.flatten(), y.flatten()
        if mic_params is None:
            mic_params = {"alpha": ALPHA, "c": C}
        self.mic_params = mic_params

    def cal_assoc(self):
        mine = MINE(**self.mic_params)
        mine.compute_score(self.x, self.y)
        return mine.mic()


class RefinedMaximalInfoCoeff(object):
    """RMIC成对检验"""
    
    def __init__(self, x: np.ndarray, y: np.ndarray, mic_params: dict = None):
        self.x, self.y = x.flatten(), y.flatten()
        self.N = len(self.x)
        if mic_params is None:
            mic_params = {"alpha": ALPHA, "c": C}
        self.mic_params = mic_params
    
    def cal_assoc(self, rebase: bool = True):
        mine = MINE(**self.mic_params)
        mine.compute_score(self.x + EPS * np.random.random(self.N), self.y + EPS * np.random.random(self.N))
        mic = mine.mic()
        
        if not rebase:
            return mic
        base_mic = self._cal_base_mic(len(self.x))
        return (mic - base_mic) / (1.0 - base_mic)
        # return (mic - base_mic) / (1.0 - base_mic) if mic > base_mic else 0.0
    
    @staticmethod
    def _cal_base_mic(n):
        """经过数值实验测试所得"""
        return 0.9893 * np.power(n, -0.292)
        
        # B = ALPHA
        # return (0.0342 - 0.1382 * B) * np.log(n) + 1.6065 * B - 0.4814
    
        # return -0.057 * np.log(n) + 0.5134
        # return model.predict(np.array([[n]]))[0]